//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DBG.H
//
//      DEBUG definitions.
//
//      NOTE:  you may ask why all this mess?  Why not just use
//      MyDebugPrintf() directly without the hassel of DBOUT(), 
//      DBOUT1(), etc.  You can use MyDebugPrintf() directly, but
//      then you can't turn them off and on dynamically at run time.
//
//      The advantange of using the DBOUT macros is that you can
//      break into the debugger at any time and set the DebugFlags
//      variable to what you want to see at the time.
//
//      Another advantage is the ability to enable the DebugThisRoutine
//      definition at the beginning of each routine to turn on all of
//      the DBOUTs in the selected routines.  Of course, this requires
//      recompilation, but it's better than nothing.
//
//      Another advantage is consistancy across the user mode (ring-3)
//      and kernel mode (ring-0) boundary.  You don't have to remember
//      how to use Output_Debug_String(), _Debug_Printf_Service(),
//      and KdPrint().  That is taken care of right here in a uniform way.
//
//      DEFINITIONS TO BE SET BEFORE INCLUDING THIS FILE:
//
//      Define the following if you want to use the DBOUTs to be inlcuded
//      into your code (otherwise they are NOPs):
//
//          DEBUG
//
//      Define one of these to differentiate between DRIVER, 
//      DLL or APPLICATION:
//
//          _DBG_DRV_
//          _DBG_DLL_
//          _DBG_APP_
//
//      If you are writing a NT driver or WDM driver, you should
//      also define the following in addition to _DBG_DRV_:
//
//          _DBG_DRV_NT_OR_WDM_
//
//      If you are writing a Linux app of driver, you should
//      also define the following in addition to _DBG_XXX_:
//
//          _DBG_LINUX_
//
//      You must specify which file is you main module by defining
//      one of the following:
//
//          _DBG_MAIN_MODULE_
//          _DBG_DRV_MAIN_MODULE_
//          _DBG_ALL_MAIN_MODULE_
//          _DBG_APP_MAIN_MODULE_
//
//      NOTE:  The use of _DBG_XXX_MAIN_MODULE_ over _DBG_MAIN_MODULE_
//      is that you can differentiate between you driver, dll and 
//      application and have different debug settings for each.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// Set the platform
// NOTE:  Assuem Windows, unless LINUX is specified
#ifndef _DBG_LINUX_
#define _DBG_WINDOWS_
#endif

#ifdef _DBG_WINDOWS_
#pragma message ("DBG.H included")
#endif

#ifndef DWORD
#define DWORD unsigned long
#endif

// Check if the proper definitions are in place
#ifndef _DBG_DRV_
    #ifndef _DBG_DLL_
        #ifndef _DBG_APP_
            #error "Need to define _DBG_DRV_, _DBG_DLL_ or _DBG_APP_!"
        #else
            #ifdef _DBG_WINDOWS_
            #pragma message("_DBG_APP_")
            #endif
        #endif
    #else
        #ifdef _DBG_WINDOWS_
        #pragma message("_DBG_DLL_")
        #endif
        #ifdef _DBG_APP_
            #ifdef _DBG_WINDOWS_
            #pragma message("_DBG_APP_")
            #endif
            #error "Cannot define _DBG_APP_ and _DBG_DLL_!"
        #endif
    #endif
#else
    #ifdef _DBG_WINDOWS_
    #pragma message("_DBG_DRV_")
    #endif
    #ifdef _DBG_APP_
        #ifdef _DBG_WINDOWS_
        #pragma message("_DBG_APP_")
        #endif
        #error "Cannot define _DBG_APP_ and _DBG_DRV_!"
    #else
        #ifdef _DBG_DLL_
            #ifdef _DBG_WINDOWS_
            #pragma message("_DBG_DLL_")
            #endif
            #error "Cannot define _DBG_DLL_ and _DBG_DRV_!"
        #endif
    #endif
#endif

// Make sure _DBG_MAIN_MODULE_ gets defined
#ifdef _DBG_DRV_MAIN_MODULE_
    #ifdef _DBG_WINDOWS_
    #pragma message("_DBG_DRV_MAIN_MODULE_")
    #endif
    #ifndef _DBG_MAIN_MODULE_
        #define _DBG_MAIN_MODULE_
    #endif
#else
    #ifdef _DBG_DLL_MAIN_MODULE_
        #ifdef _DBG_WINDOWS_
        #pragma message("_DBG_DLL_MAIN_MODULE_")
        #endif
        #ifndef _DBG_MAIN_MODULE_
            #define _DBG_MAIN_MODULE_
        #endif
    #else
        #ifdef _DBG_APP_MAIN_MODULE_
            #ifdef _DBG_WINDOWS_
            #pragma message("_DBG_APP_MAIN_MODULE_")
            #endif
            #ifndef _DBG_MAIN_MODULE_
                #define _DBG_MAIN_MODULE_
            #endif
        #endif
    #endif
#endif

#ifndef _DBG_H_

// WE HAVEN'T INCLUDED THIS FILE YET!

#define _DBG_H_

// RUN-TIME Debug Flag definitions

// HIGH FREQUENCY
#define DBG_HF_ENTRY            0x00000001
#define DBG_HF_EXIT             0x00000002
#define DBG_HF_ENTRY_PARAMS     0x00000004
#define DBG_HF_EXIT_PARAMS      0x00000008
#define DBG_HF_INTERNAL         0x00000010
// Your improvements here
//#define DBG_HF_               0x00000020
//#define DBG_HF_               0x00000040
//#define DBG_HF_               0x00000080

// MEDIUM FREQUENCY
#define DBG_MF_ENTRY            0x00000100
#define DBG_MF_EXIT             0x00000200
#define DBG_MF_ENTRY_PARAMS     0x00000400
#define DBG_MF_EXIT_PARAMS      0x00000800
#define DBG_MF_INTERNAL         0x00001000
// Your improvements here
//#define DBG_MF_               0x00002000
//#define DBG_MF_               0x00004000
//#define DBG_MF_               0x00008000

// LOW FREQUENCY
#define DBG_LF_ENTRY            0x00010000
#define DBG_LF_EXIT             0x00020000
#define DBG_LF_ENTRY_PARAMS     0x00040000
#define DBG_LF_EXIT_PARAMS      0x00080000
#define DBG_LF_INTERNAL         0x00100000
// Your improvements here
//#define DBG_LF_               0x00200000
//#define DBG_LF_               0x00400000
//#define DBG_LF_               0x00800000

// MISC
#define DBG_ALWAYS              0x80000000
#define DBG_CUR_DB              0x40000000
// Your improvements here
#define DBG_READ_REGS           0x20000000  // reads hw registers
#define DBG_WRITE_REGS          0x10000000  // writes to hw regs
//#define DBG_                  0x08000000
//#define DBG_                  0x04000000
//#define DBG_                  0x02000000
//#define DBG_                  0x01000000

// PREDEFINED COMBINATIONS
#define DBG_ALL                 0xFFFFFFFF
#define DBG_NONE                0
#define DBG_ENTRY               DBG_HF_ENTRY | DBG_MF_ENTRY | DBG_LF_ENTRY
#define DBG_EXIT                DBG_HF_EXIT | DBG_MF_EXIT | DBG_LF_EXIT
#define DBG_ERROR               DBG_ALWAYS
#define DBG_DEFAULT             DBG_ALWAYS | DBG_ERROR | DBG_CUR_DB
#define DBG_REGS                DBG_READ_REGS | DBG_WRITE_REGS

#ifdef DEBUG

    // THIS IS DEBUG!
    #ifdef _DBG_WINDOWS_
    #pragma message("DEBUG DEFINED!")
    #endif

    // The main trick -- all DBOUTs use DEBUGFLAGS as their
    // basis for activation.  Defining DEBUGFLAGS as DebugThisRoutine
    // will make all of the DBOUTs in the routine active for a
    // particular build.
    #define DEBUGFLAGS DebugFlags

    // Driver debug flags -- keep separate from your dlls and apps
    #ifdef _DBG_DRV_MAIN_MODULE_

        #ifdef _DBG_WINDOWS_
        #pragma message("DRIVER MAIN MODULE!")
        #endif

        DWORD DebugFlags =          DBG_DEFAULT;
        DWORD DebugThisRoutine =    DBG_ALL;

    #else

        // Dll debug flags -- keep separate from your drvs and apps
        #ifdef _DBG_DLL_MAIN_MODULE_

            #ifdef _DBG_WINDOWS_
            #pragma message("LIBRARY MAIN MODULE!")
            #endif

            DWORD DebugFlags =          DBG_DEFAULT;
            DWORD DebugThisRoutine =    DBG_ALL;

        #else

            // App debug flags -- keep separate from your drvs and dlls
            #ifdef _DBG_APP_MAIN_MODULE_

                #ifdef _DBG_WINDOWS_
                #pragma message("APP MAIN MODULE!")
                #endif

                DWORD DebugFlags =          DBG_DEFAULT;
                DWORD DebugThisRoutine =    DBG_ALL;

            #else

                // Common debug flags -- keep same for your drvs, dlls and apps
                #ifdef _DBG_MAIN_MODULE_

                    #ifdef _DBG_WINDOWS_
                    #pragma message("GENERAL MODULE!")
                    #endif

                    DWORD DebugFlags =          DBG_DEFAULT;
                    DWORD DebugThisRoutine =    DBG_ALL;

                #else

                    // If not a main modules of DRV, DLL or APP, then debug flags var is externed
                    #ifdef _DBG_WINDOWS_
                    #pragma message("EXTERNAL MODULE!")
                    #endif

                    extern DWORD DebugFlags;
                    extern DWORD DebugThisRoutine;

                #endif

            #endif

        #endif

    #endif

    // DBOUT definitions
    #define MY_DBG_MSG_SIZE 500
    #ifdef _DBG_WINDOWS_
    #pragma warning (disable:4005)
    #endif
    #ifdef _DBG_DRV_
        #define DBBREAK {if(DEBUGFLAGS&(DBG_ERROR)){_asm int 1}}
    #else
        #define DBBREAK {if(DEBUGFLAGS&(DBG_ERROR)){_asm int 3}}
    #endif
    #define DBOUT(f,s) {if(DEBUGFLAGS&(f)) MyDebugPrintf(s);}
    #define DBOUT1(f,s,p1) {if(DEBUGFLAGS&(f)) MyDebugPrintf(s,p1);}
    #define DBOUT2(f,s,p1,p2) {if(DEBUGFLAGS&(f)) MyDebugPrintf(s,p1,p2);}
    #define DBOUT3(f,s,p1,p2,p3) {if(DEBUGFLAGS&(f)) MyDebugPrintf(s,p1,p2,p3);}
    #define DBOUT4(f,s,p1,p2,p3,p4) {if(DEBUGFLAGS&(f)) MyDebugPrintf(s,p1,p2,p3,p4);}
    #define DBOUT5(f,s,p1,p2,p3,p4,p5) {if(DEBUGFLAGS&(f)) MyDebugPrintf(s,p1,p2,p3,p4,p5);}

    // VxDs use _Debug_Printf_Service()
    // NT and WDM drivers use KdPrint()
    // DLLs and APPs use OutputDebugString()
    // Linux Drivers us printk()
    // Linux Apps us printf()

    // NOTE:  each module will have it's own copy of this routine.
    // If that is undesirable, then use the _XXX_MAIN_MODULE_ definitions
    // to make one public definition and extern the others.
    static void MyDebugPrintf (char *pszfmt, ...)
    {
        #ifdef _DBG_DRV_

            #ifndef _DBG_DRV_NT_OR_WDM_

                #ifdef _DBG_WINDOWS_

                    // VxD
                    #ifdef _DBG_WINDOWS_
                    #pragma message("DRIVER = VXD!")
                    #endif

                    // Comment this out if you can include VMM.h and don't trust
                    // my kludgy in-line emits below.
                    #define DEBUG_PRINTF_IN_LINE

                    #ifdef DEBUG_PRINTF_IN_LINE

                        // THIS IS IN LINE
                        // USEFUL for when you don't have the convenience of
                        // linking with VxDWraps.CLB or including VMM.H.
                        _asm
                        {
                            lea eax,(pszfmt + 4)
                            push    eax
                            push    pszfmt

                            _emit   0xCD;       // int 20
                            _emit   0x20
                            _emit   0x2D;       // function 0x2D010100 (_Debug_Printf_Service)
                            _emit   0x01
                            _emit   0x01
                            _emit   0x00

                            add esp, 2*4
                        }

                    #else

                        // This is the way you do it if can include VMM.H

                        _asm
                        {
                            lea     eax,(pszfmt + 4)
                            push    eax
                            push    pszfmt

                        }
                        VMMCall(_Debug_Printf_Service)
                        _asm
                        {
                            add esp, 2*4
                        }

                    #endif

                #else

                        // LINUX Driver
                        #ifdef _DBG_WINDOWS_
                        #pragma message("DRIVER = LINUX!")
                        #endif

                        static char msg[MY_DBG_MSG_SIZE];
                        static char locfmt[MY_DBG_MSG_SIZE];
                        DWORD *p;

                        p = (DWORD *)&pszfmt;
                        p++;
                        sprintf(locfmt, "<1>%s", pszfmt, *p, *(p+1), *(p+2), *(p+3), *(p+4));
                        sprintf(msg, locfmt, *p, *(p+1), *(p+2), *(p+3), *(p+4));
                        printk (msg);

                #endif

            #else

                // NT or WDM driver
                #ifdef _DBG_WINDOWS_
                #pragma message("DRIVER = NT OR WDM!")
                #endif

                DWORD *p;

                p = (DWORD *)&pszfmt;
                p++;
                KdPrint((pszfmt, *p, *(p+1), *(p+2), *(p+3), *(p+4)));

            #endif

        #else

            #ifdef _DBG_WINDOWS_

                // DLL or APP
                #ifdef _DBG_WINDOWS_
                #pragma message("APP/LIB = WINDOWS!")
                #endif

                char static msg[MY_DBG_MSG_SIZE];
                DWORD *p;

                p = (DWORD *)&pszfmt;
                p++;
                sprintf(msg, pszfmt, *p, *(p+1), *(p+2), *(p+3), *(p+4));
                OutputDebugString (msg);

            #else

                // LINUX LIBRARY or APP
                #ifdef _DBG_WINDOWS_
                #pragma message("APP/LIB = LINUX!")
                #endif

                DWORD *p;

                p = (DWORD *)&pszfmt;
                p++;
                printf (pszfmt, *p, *(p+1), *(p+2), *(p+3), *(p+4));
//                fprintf (stderr, pszfmt, *p, *(p+1), *(p+2), *(p+3), *(p+4));
//              syslog (LOG_ERR|LOG_USER, pszfmt, *p, *(p+1), *(p+2), *(p+3), *(p+4));

            #endif

        #endif
    }

    #ifdef _DBG_MAIN_MODULE_

        // Need to call this from DriverEntry(), Main(), WinMain() or LibMain()
        void DBGInitialize (DWORD dbgFlags)
        {
            // For initial development, turn on everything
            DebugFlags = dbgFlags;
            //DebugFlags = DBG_ALL;

            // Set a break here and edit the DebugFlags variable
//          DBOUT (DBG_ALWAYS, "\r\n");
//          DBOUT (DBG_ALWAYS, "\r\n");
//          DBOUT (DBG_ALWAYS, "*********************************************************************\r\n");
//          DBOUT (DBG_ALWAYS, "*********************************************************************\r\n");
//          DBOUT (DBG_ALWAYS, "\r\n");
//          DBOUT (DBG_ALWAYS, "Edit the DebugFlags variable to enable run-time debug trace out flags\r\n");
            DBOUT1 (DBG_ALWAYS, "&DebugFlags = %08X\r\n", &DebugFlags);
            DBOUT1 (DBG_ALWAYS, "DebugFlags = %08X\r\n", DebugFlags);
//          DBOUT (DBG_ALWAYS, "\r\n");
            DBOUT2 (DBG_ALWAYS, "DBG_HF_ENTRY          = %08X %s\r\n", DBG_HF_ENTRY,        DebugFlags&DBG_HF_ENTRY?"ON":"");
            DBOUT2 (DBG_ALWAYS, "DBG_HF_EXIT           = %08X %s\r\n", DBG_HF_EXIT,         DebugFlags&DBG_HF_EXIT?"ON":"");
            DBOUT2 (DBG_ALWAYS, "DBG_HF_ENTRY_PARAMS   = %08X %s\r\n", DBG_HF_ENTRY_PARAMS, DebugFlags&DBG_HF_ENTRY_PARAMS?"ON":"");
            DBOUT2 (DBG_ALWAYS, "DBG_HF_EXIT_PARAMS    = %08X %s\r\n", DBG_HF_EXIT_PARAMS,  DebugFlags&DBG_HF_EXIT_PARAMS?"ON":"");
            DBOUT2 (DBG_ALWAYS, "DBG_HF_INTERNAL       = %08X %s\r\n", DBG_HF_INTERNAL,     DebugFlags&DBG_HF_INTERNAL?"ON":"");
            DBOUT2 (DBG_ALWAYS, "DBG_MF_ENTRY          = %08X %s\r\n", DBG_MF_ENTRY,        DebugFlags&DBG_MF_ENTRY?"ON":"");
            DBOUT2 (DBG_ALWAYS, "DBG_MF_EXIT           = %08X %s\r\n", DBG_MF_EXIT,         DebugFlags&DBG_MF_EXIT?"ON":"");
            DBOUT2 (DBG_ALWAYS, "DBG_MF_ENTRY_PARAMS   = %08X %s\r\n", DBG_MF_ENTRY_PARAMS, DebugFlags&DBG_MF_ENTRY_PARAMS?"ON":"");
            DBOUT2 (DBG_ALWAYS, "DBG_MF_EXIT_PARAMS    = %08X %s\r\n", DBG_MF_EXIT_PARAMS,  DebugFlags&DBG_MF_EXIT_PARAMS?"ON":"");
            DBOUT2 (DBG_ALWAYS, "DBG_MF_INTERNAL       = %08X %s\r\n", DBG_MF_INTERNAL,     DebugFlags&DBG_MF_INTERNAL?"ON":"");
            DBOUT2 (DBG_ALWAYS, "DBG_LF_ENTRY          = %08X %s\r\n", DBG_LF_ENTRY,        DebugFlags&DBG_LF_ENTRY?"ON":"");
            DBOUT2 (DBG_ALWAYS, "DBG_LF_EXIT           = %08X %s\r\n", DBG_LF_EXIT,         DebugFlags&DBG_LF_EXIT?"ON":"");
            DBOUT2 (DBG_ALWAYS, "DBG_LF_ENTRY_PARAMS   = %08X %s\r\n", DBG_LF_ENTRY_PARAMS, DebugFlags&DBG_LF_ENTRY_PARAMS?"ON":"");
            DBOUT2 (DBG_ALWAYS, "DBG_LF_EXIT_PARAMS    = %08X %s\r\n", DBG_LF_EXIT_PARAMS,  DebugFlags&DBG_LF_EXIT_PARAMS?"ON":"");
            DBOUT2 (DBG_ALWAYS, "DBG_LF_INTERNAL       = %08X %s\r\n", DBG_LF_INTERNAL,     DebugFlags&DBG_LF_INTERNAL?"ON":"");
            DBOUT2 (DBG_ALWAYS, "DBG_ALWAYS            = %08X %s\r\n", DBG_ALWAYS,          DebugFlags&DBG_ALWAYS?"ON":"");
            DBOUT2 (DBG_ALWAYS, "DBG_CUR_DB            = %08X %s\r\n", DBG_CUR_DB,          DebugFlags&DBG_CUR_DB?"ON":"");
            DBOUT2 (DBG_ALWAYS, "DBG_READ_REGS         = %08X %s\r\n", DBG_READ_REGS,       DebugFlags&DBG_READ_REGS?"ON":"");
            DBOUT2 (DBG_ALWAYS, "DBG_WRITE_REGS        = %08X %s\r\n", DBG_WRITE_REGS,      DebugFlags&DBG_WRITE_REGS?"ON":"");
//          DBOUT (DBG_ALWAYS, "\r\n");
            #ifdef _DBG_WINDOWS_
            _asm int 3;
            #endif
//          DBOUT (DBG_ALWAYS, "*********************************************************************\r\n");
//          DBOUT (DBG_ALWAYS, "*********************************************************************\r\n");
//          DBOUT (DBG_ALWAYS, "\r\n");
//          DBOUT (DBG_ALWAYS, "\r\n");

        }

        #ifndef MyIsPrint
        #define MyIsPrint(c) (c>=' ' && c<='~')
        #endif
        void DBGDumpBuffer (char *caption, unsigned char *buffer, int bufferLen, int bytesPerUnit, int unitsPerLine)
        {
            int i;
            int j;
            __u8 *p;
            __u16 *p16;
            __u32 *p32;
            int unitsLeft;
            char lineStr[200];

            DBOUT1 (DBG_ALWAYS, "%s\r\n", caption);

            // Dump according to granularity
            switch (bytesPerUnit)
            {
            // DUMP __u8
            default:
            case sizeof(__u8):
                for (p=buffer, i=0; i < bufferLen; )
                {
                    lineStr[0] = '\0';
                    sprintf (lineStr, "%08X: ", (int)p);
                    unitsLeft = bufferLen - i;
                    if (unitsLeft >= unitsPerLine)
                    {
                        for (j=0; j<unitsPerLine; j++)
                        {
                            sprintf (lineStr+strlen(lineStr), "%02X ", *(p+j));
                        }
                        sprintf (lineStr+strlen(lineStr), "    ");
                        for (j=0; j<unitsPerLine; j++)
                        {
                            sprintf (lineStr+strlen(lineStr), "%c", MyIsPrint(*(p+j))?*(p+j):'.');
                        }
                        i += unitsPerLine;
                        p += unitsPerLine;
                    }
                    else
                    {
                        for (j=0; j<unitsLeft; j++)
                        {
                            sprintf (lineStr+strlen(lineStr), "%02X ", *(p+j));
                        }
                        for (; j<unitsPerLine; j++)
                        {
                            sprintf (lineStr+strlen(lineStr), "   ");
                        }
                        sprintf (lineStr+strlen(lineStr), "    ");
                        for (j=0; j<unitsLeft; j++)
                        {
                            sprintf (lineStr+strlen(lineStr), "%c", MyIsPrint(*(p+j))?*(p+j):'.');
                        }
                        i += unitsLeft;
                        p += unitsLeft;
                    }
                    DBOUT1 (DBG_ALWAYS, "%s\n\r", lineStr);
                }
                break;

            // DUMP __u16
            case sizeof(__u16):
                for (p16=(__u16 *)buffer, i=0; i < bufferLen; )
                {
                    lineStr[0] = '\0';
                    sprintf (lineStr, "%08X: ", (int)p16);
                    unitsLeft = (bufferLen - i)/sizeof(__u16);
                    if (unitsLeft >= unitsPerLine)
                    {
                        for (j=0; j<unitsPerLine; j++)
                        {
                            sprintf (lineStr+strlen(lineStr), "%04X ", *(p16+j));
                        }
                        i += (unitsPerLine * sizeof(__u16));
                        p16 += unitsPerLine;
                    }
                    else
                    {
                        for (j=0; j<unitsLeft; j++)
                        {
                            sprintf (lineStr+strlen(lineStr), "%04X ", *(p16+j));
                        }
                        i += (unitsLeft * sizeof(__u16));
                        p16 += unitsLeft;
                    }
                    DBOUT1 (DBG_ALWAYS, "%s\n\r", lineStr);
                }
                break;

            // DUMP __u32
            case sizeof(__u32):
                for (p32=(__u32 *)buffer, i=0; i < bufferLen; )
                {
                    lineStr[0] = '\0';
                    sprintf (lineStr, "%08X: ", (int)p32);
                    unitsLeft = (bufferLen - i)/sizeof(__u32);
                    if (unitsLeft >= unitsPerLine)
                    {
                        for (j=0; j<unitsPerLine; j++)
                        {
                            sprintf (lineStr+strlen(lineStr), "%08X ", *(p32+j));
                        }
                        i += (unitsPerLine * sizeof(__u32));
                        p32 += unitsPerLine;
                    }
                    else
                    {
                        for (j=0; j<unitsLeft; j++)
                        {
                            sprintf (lineStr+strlen(lineStr), "%08X ", *(p32+j));
                        }
                        i += (unitsLeft * sizeof(__u32));
                        p32 += unitsLeft;
                    }
                    DBOUT1 (DBG_ALWAYS, "%s\n\r", lineStr);
                }
                break;
            }
        }

    #else
        void DBGDumpBuffer (char *caption, unsigned char *buffer, int bufferLen, int bytesPerUnit, int unitsPerLine);
    #endif

#else

    // THIS IS NON-DEBUG
    // ALL DBOUTs are vapor
    #ifdef _DBG_WINDOWS_
    #pragma message("DEBUG NOT DEFINED!")
    #endif

    #define DBBREAK
    #define DBOUT(f,s)
    #define DBOUT1(f,s,p1)
    #define DBOUT2(f,s,p1,p2)
    #define DBOUT3(f,s,p1,p2,p3)
    #define DBOUT4(f,s,p1,p2,p3,p4)
    #define DBOUT5(f,s,p1,p2,p3,p4,p5)
    #ifdef _DBG_MAIN_MODULE_
        #define DBGInitialize(dbgFlags)
        #define DBGDumpBuffer(caption,buffer,bufferLen,bytesPreUnit,unitsPerLine)
    #endif

#endif

// WE HAVE ALREADY INCLUDED THIS FILE -- WE'RE OUT OF HERE!

#endif
